@@ -743,6 +743,9 @@ let to_buffer ?separator ?backslash_escape ?excel_tricks buf =
        method close_out () = ()
      end)
 
+let close_out oc =
+  oc.out_chan#close_out()
+
 let rec really_output oc s ofs len =
   let w = oc.out_chan#output s ofs len in
   if w < len then really_output oc s (ofs+w) (len-w)
@@ -854,7 +857,7 @@ let save ?separator ?backslash_escape ?excel_tricks fname t =
   let ch = open_out fname in
   let csv = to_channel ?separator ?backslash_escape ?excel_tricks ch in
   output_all csv t;
-  close_out ch
+  Pervasives.close_out ch
 
 (*
  * Reading rows with headers
@@ -1099,18 +1102,13 @@ let to_array csv =
 let of_array csv =
   List.map Array.to_list (Array.to_list csv)
 
+let rec combine ~header row = match header, row with
+  | [], _ -> []
+  | _, [] -> List.map (fun h -> (h, "")) header
+  | h0 :: h, x :: r -> (h0, x) :: combine ~header:h r
+
 let associate header data =
-  let nr_cols = List.length header in
-  let rec trunc = function
-    | 0, _ -> []
-    | n, [] -> "" :: trunc (n-1, [])
-    | n, (x :: xs) -> x :: trunc (n-1, xs)
-  in
-  List.map (
-    fun row ->
-      let row = trunc (nr_cols, row) in
-      List.combine header row
-  ) data
+  List.map (fun row -> combine header row) data
 
 let map ~f csv =
   List.map (fun row -> List.map (fun el -> f el) row) csv
